@@ -4,13 +4,17 @@ import importlib
 import logging
 import operator
 from collections import Counter
-from dataclasses import dataclass
+from dataclasses import dataclass, field
 from datetime import datetime
 from types import ModuleType
 from typing import Any, Dict, List, Literal, Optional, Protocol, Tuple, Union
+from numbers import Number
+from .iso.checks import ISO
 
 from colorama import Fore, Style  # type: ignore
 from toolz import valfilter  # type: ignore
+import numpy as np
+
 logger = logging.getLogger("cuallee")
 
 # Verify Libraries Available
@@ -37,6 +41,13 @@ try:
 except:
     logger.debug(Fore.RED + "[KO]" + Fore.WHITE + " Snowpark")
 
+try:
+    from duckdb import DuckDBPyConnection as duckdb_dataframe  # type: ignore
+
+    logger.debug(Fore.GREEN + "[OK]" + Fore.WHITE + " DuckDB")
+except:
+    logger.debug(Fore.RED + "[KO]" + Fore.WHITE + " DuckDB")
+
 
 logger.debug(Style.RESET_ALL)
 
@@ -70,13 +81,21 @@ class Rule:
     value: Optional[Any]
     data_type: CheckDataType
     coverage: float = 1.0
+    options: Union[List[Tuple], None] = None
     status: Union[str, None] = None
 
+    @property
+    def settings(self) -> dict:
+        return dict(self.options)
+
     @property
     def key(self):
         return (
             hashlib.blake2s(
-                bytes(f"{self.method}{self.column}{self.value}{self.coverage}", "utf-8")
+                bytes(
+                    f"{self.method}{self.column}{self.value}{self.options}{self.coverage}",
+                    "utf-8",
+                )
             )
             .hexdigest()
             .upper()
@@ -122,9 +141,11 @@ class Check:
         level: Union[CheckLevel, int],
         name: str,
         execution_date: datetime = datetime.today(),
+        table_name: str = None,
     ):
         """A container of data quality rules."""
         self._rule: Dict[str, Rule] = {}
+        # TODO: Should be a compute engine protocol
         self.compute_engine: ModuleType
 
         if isinstance(level, int):
@@ -136,9 +157,14 @@ class Check:
         self.date = execution_date
         self.rows = -1
         self.config: Dict[str, str] = {}
+        self.table_name = table_name
+        self.iso = ISO(self)
 
     def __repr__(self):
-        return f"Check(level:{self.level}, desc:{self.name}, rules:{self.sum})"
+        standard = f"Check(level:{self.level}, desc:{self.name}, rules:{self.sum})"
+        if self.table_name:
+            standard += f" / table:{self.table_name}"
+        return standard
 
     @property
     def sum(self):
@@ -294,6 +320,11 @@ class Check:
         Rule("has_mean", column, value, CheckDataType.NUMERIC) >> self._rule
         return self
 
+    def has_sum(self, column: str, value: float):
+        """Validation of a sum of all values of a column"""
+        Rule("has_sum", column, value, CheckDataType.NUMERIC) >> self._rule
+        return self
+
     def is_between(self, column: str, value: Tuple[Any], pct: float = 1.0):
         """Validation of a column between a range"""
         Rule("is_between", column, value, CheckDataType.AGNOSTIC, pct) >> self._rule
@@ -319,21 +350,19 @@ class Check:
         return self.is_contained_in(column, value, pct)
 
     def has_percentile(
-        self,
-        column: str,
-        value: float,
-        percentile: float,
-        precision: int = 10000,
-        pct: float = 1.0,
+        self, column: str, value: float, percentile: float, precision: int = 10000
     ):
         """Validation of a column percentile value"""
         (
             Rule(
                 "has_percentile",
                 column,
-                (value, percentile, precision),
+                value,
                 CheckDataType.NUMERIC,
-                pct,
+                options=[
+                    tuple(["percentile", percentile]),
+                    tuple(["precision", precision]),
+                ],
             )
             >> self._rule
         )
@@ -356,11 +385,7 @@ class Check:
         return self
 
     def has_max_by(
-        self,
-        column_source: str,
-        column_target: str,
-        value: Union[float, str],
-        pct: float = 1.0,
+        self, column_source: str, column_target: str, value: Union[float, str]
     ):
         """Validation of a column value based on another column maximum"""
         (
@@ -375,11 +400,7 @@ class Check:
         return self
 
     def has_min_by(
-        self,
-        column_source: str,
-        column_target: str,
-        value: Union[float, str],
-        pct: float = 1.0,
+        self, column_source: str, column_target: str, value: Union[float, str]
     ):
         """Validation of a column value based on another column minimum"""
         (
@@ -393,9 +414,7 @@ class Check:
         )
         return self
 
-    def has_correlation(
-        self, column_left: str, column_right: str, value: float, pct: float = 1.0
-    ):
+    def has_correlation(self, column_left: str, column_right: str, value: float):
         """Validates the correlation between 2 columns with some tolerance"""
         (
             Rule(
@@ -416,7 +435,13 @@ class Check:
     def has_entropy(self, column: str, value: float, tolerance: float = 0.01):
         """Validation for entropy calculation on continuous values"""
         (
-            Rule("has_entropy", column, (value, tolerance), CheckDataType.AGNOSTIC)
+            Rule(
+                "has_entropy",
+                column,
+                value,
+                CheckDataType.AGNOSTIC,
+                options=[tuple(["tolerance", tolerance])],
+            )
             >> self._rule
         )
         return self
@@ -481,6 +506,27 @@ class Check:
         (Rule("is_daily", column, value, CheckDataType.DATE, pct) >> self._rule)
         return self
 
+    def has_workflow(
+        self,
+        column_group: str,
+        column_event: str,
+        column_order: str,
+        edges: List[Tuple[str]],
+        pct: float = 1.0,
+    ):
+        """Validates events in a group clause with order, followed a specific sequence. Similar to adjacency matrix validation"""
+        (
+            Rule(
+                "has_workflow",
+                [column_group, column_event, column_order],
+                edges,
+                CheckDataType.AGNOSTIC,
+                pct,
+            )
+            >> self._rule
+        )
+        return self
+
     def validate(self, dataframe: Any):
         """Compute all rules in this check for specific data frame"""
 
@@ -504,6 +550,11 @@ class Check:
         ):
             self.compute_engine = importlib.import_module("cuallee.snowpark_validation")
 
+        elif "duckdb_dataframe" in globals() and isinstance(
+            dataframe, duckdb_dataframe
+        ):
+            self.compute_engine = importlib.import_module("cuallee.duckdb_validation")
+
         assert self.compute_engine.validate_data_types(
             self.rules, dataframe
         ), "Invalid data types between rules and dataframe"
